from engineio_client.emitter import Emitter
from engineio_client.client import Client as EngineIOClient
from engineio_client.utils import format_long
from .parser import Parser, ParserException, PacketException
from .socket import Socket
import gevent

import logging
logger = logging.getLogger(__name__)

class Manager(Emitter):
    def __init__(self, scheme, hostname, port, reconnection=True, reconnection_delay=3.0,
        auto_connect=True, parser=None, **kwargs):
        """
        scheme:                 URI scheme
        hostname:               Hostname of the server
        port:                   Port of the server
        reconnection:           Automatically reconnect after disconnect
        reconnection_delay:     How long to wait between reconnections float in seconds
        auto_connect:           Automatically connect at creation
        """
        super(Manager, self).__init__()

        self.scheme = scheme
        self.hostname = hostname
        self.port = port
        self.reconnection = reconnection
        self.reconnection_delay = reconnection_delay
        self.auto_connect = auto_connect
        self.parser = parser or Parser()
        self.engine_kwargs = kwargs
        self.engine_kwargs.setdefault('path', '/socket.io')
        
        self.state = 'closed'
        self.sockets = set()
        self.engine = None
        self.reconnecting = False
        self.skipping_reconnect = False
        self.reconnect_task = False

        if self.auto_connect:
            self.connect()

    def setup_engine_handlers(self):
        self.engine.on('open', self.handle_open)
        self.engine.on('error', self.handle_error)
        self.engine.on('message', self.handle_message)
        self.engine.on('close', self.handle_close)

    def cleanup_engine_handlers(self):
        self.engine.off('open', self.handle_open)
        self.engine.off('error', self.handle_error)
        self.engine.off('message', self.handle_message)
        self.engine.off('close', self.handle_close)

    def connect(self):
        if self.state in ['opening', 'open']:
            return

        self.state = 'opening'
        self.skip_reconnect(False)

        self.engine = EngineIOClient(self.scheme, self.hostname, self.port, **self.engine_kwargs)
        self.setup_engine_handlers()
        self.engine.open()

    def disconnect(self):
        logger.debug("Disconnecting")
        self.reconnecting = False
        self.skip_reconnect(True)
        self.state = 'closing'
        if self.engine:
            self.engine.close()

    def send_packet(self, packet):
        logger.debug(format_long("Sending packet: %s", packet))
        items = self.parser.encode(packet)
        binary = False
        for item in items:
            self.engine.send(item, binary)
            binary = True   # Attachments are always considered binary

    def socket(self, namespace):
        return Socket(namespace, self)

    def attach_socket(self, socket):
        self.sockets.add(socket)

    def detach_socket(self, socket):
        logger.debug("Detach socket")
        self.sockets.discard(socket)
        if not self.sockets:
            self.disconnect()

    @property
    def id(self):
        return self.engine.id

    def handle_open(self):
        if self.state == 'opening':
            logger.debug("Connected")
            self.state = 'open'
            self.emit('open')
            if self.reconnecting:
                self.emit('reconnect')
                self.reconnecting = False

    def handle_error(self, error):
        if self.state == 'opening':

            logger.warning("Connect error")
            self.state = 'closed'
            self.emit('connect_error', error)

            if self.reconnecting:
                logger.warning("Reconnect error")
                self.emit('reconnect_error', error)
                self.reconnecting = False
                self.reconnect()
            elif self.reconnection:
                self.reconnect()

        else:
            logger.warning("Error: %s", error)
            self.emit('error', error)

    def handle_message(self, message):
        if self.state != 'open':
            return
        
        try:
            packet = self.parser.decode(message)
            if packet:
                self.handle_packet(packet)
        except (ParserException, PacketException) as e:
            logger.warning("Received invalid message or sequence, ignoring: %s", e)

    def handle_packet(self, packet):
        if self.state != 'open':
            return

        logger.debug(format_long("Received packet: %s", packet))
        self.emit('packet', packet)

    def handle_close(self):
        logger.debug("Handle close")
        if self.state != 'closing' and self.state != 'open':
            return

        logger.debug("Disconnected")
        self.state = 'closed'
        self.emit('close')

        self.cleanup_engine_handlers()

        if self.reconnection:
            self.reconnect()

    def skip_reconnect(self, value):
        self.skipping_reconnect = value
        if self.skipping_reconnect:
            self.stop_task(self.reconnect_task)

    def do_reconnect(self):
        if self.state in ['opening', 'open']:
            return

        logger.debug("Reconnect")
        self.connect()
    
    def reconnect(self):
        logger.debug("Delay reconnect")
        if self.reconnecting:
            logger.debug("Already reconnecting")
            return

        if self.skipping_reconnect:
            logger.debug("Skipping reconnect because user disconnected")
            return

        self.reconnecting = True
        self.reconnect_task = self.start_task(self.do_reconnect, delay=self.reconnection_delay)

    def start_task(self, func, delay=0, *args, **kwargs):
        return gevent.spawn_later(delay, func, *args, **kwargs)

    def stop_task(self, task):
        if task:
            task.kill(False)




#!/usr/bin/env python
# -*- coding: us-ascii -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab
#
"""
REL Test Suite - regular unittest compatible suite
Submitted by Chris Clark on 10/12/2011. Thanks, Chris!

Original code here:
http://code.google.com/r/clach04-pyeventtestfixes/source/browse/test.py

Sample usage:

    rel\test.py
    rel\test.py -v
    rel\test.py -v EventTest.test_exception EventTest.test_timeout
    ... etc.

"""

from . import rel
rel.override()

import glob
import os
import signal
import sys
import _thread
import time
import unittest

"""
# skip code, disabled due to odd raise in Cython code
try:
    if sys.version_info < (2, 3):
        # unittest2 needs 2.3+
        raise ImportError
    import unittest2
    unittest = unittest2
except ImportError:
    import unittest
    unittest2 = None
"""
unittest2 = None

# Note "setup.py develop" is preferred mechanism for build areas
possible_build_dir = glob.glob('./build/lib.*')
if possible_build_dir:
    sys.path.insert(0, possible_build_dir[0])

import event

class EventTest(unittest.TestCase):

    def skip(self, reason):
        """Skip current test because of `reason`.
       
        NOTE currently expects unittest2, and defaults to "pass"
        if not available.
       
        unittest2 does NOT work under Python 2.2.
        Could potentially use nose or py.test which has (previously)
        supported Python 2.2
          * nose
            http://python-nose.googlecode.com/svn/wiki/NoseWithPython2_2.wiki
          * py.test
            http://codespeak.net/pipermail/py-dev/2005-February/000203.html
        """
        if unittest2:
            raise unittest2.SkipTest(reason)

    def setUp(self):
        self.call_back_ran = False
        event.init()

    def test_timeout(self):
        def __timeout_cb(ev, handle, evtype, ts):
            now = time.time()
            self.call_back_ran = True
            assert int(now - ts['start']) == ts['secs'], 'timeout failed'
        ts = {'start': time.time(), 'secs': 5}
        ev = event.event(__timeout_cb, arg=ts)
        ev.add(ts['secs'])
        event.dispatch()
        self.assertTrue(self.call_back_ran, 'call back did not run')

    def test_timeout2(self):
        def __timeout2_cb(start, secs):
            self.call_back_ran = True
            dur = int(time.time() - start)
            assert dur == secs, 'timeout2 failed'
        event.timeout(5, __timeout2_cb, time.time(), 5)
        event.dispatch()
        self.assertTrue(self.call_back_ran, 'call back did not run')
   
    def test_signal(self):
        if not hasattr(signal, 'SIGUSR1'):
            self.skip('signal.SIGUSR1 missing (probably Windows)')
        else:
           
            def __signal_cb(ev, sig, evtype, arg):
                if evtype == event.EV_SIGNAL:
                    ev.delete()
                elif evtype == event.EV_TIMEOUT:
                    os.kill(os.getpid(), signal.SIGUSR1)
            event.event(__signal_cb, handle=signal.SIGUSR1,
                        evtype=event.EV_SIGNAL).add()
            event.event(__signal_cb).add(2)
            event.dispatch()

    def test_signal2(self):
        if not hasattr(signal, 'SIGUSR1'):
            self.skip('signal.SIGUSR1 missing (probably Windows)')
        else:
           
            def __signal2_cb(sig):
                if sig:
                    event.abort()
                else:
                    os.kill(os.getpid(), signal.SIGUSR1)
            event.signal(signal.SIGUSR1, __signal2_cb, signal.SIGUSR1)
            event.timeout(2, __signal2_cb)

    def test_read(self):
        def __read_cb(ev, fd, evtype, pipe):
            self.call_back_ran = True
            buf = os.read(fd, 1024)
            assert buf == 'hi niels', 'read event failed'
        pipe = os.pipe()
        event.event(__read_cb, handle=pipe[0],
                    evtype=event.EV_READ).add()
        os.write(pipe[1], 'hi niels')
        event.dispatch()
        self.assertTrue(self.call_back_ran, 'call back did not run')

    def test_read2(self):
        def __read2_cb(fd, msg):
            self.call_back_ran = True
            assert os.read(fd, 1024) == msg, 'read2 event failed'
        msg = 'hello world'
        pipe = os.pipe()
        event.read(pipe[0], __read2_cb, pipe[0], msg)
        os.write(pipe[1], msg)
        event.dispatch()
        self.assertTrue(self.call_back_ran, 'call back did not run')

    def test_exception(self):
        def __bad_cb(foo):
            raise NotImplementedError(foo)

        def real_test_exception():
            event.timeout(0, __bad_cb, 'bad callback')
            event.dispatch()

        self.assertRaises(NotImplementedError, real_test_exception)

    def test_abort(self):
        def __time_cb():
            raise NotImplementedError('abort failed!')
        event.timeout(5, __time_cb)
        event.timeout(1, event.abort)
        event.dispatch()

    def test_callback_exception(self):
        def __raise_cb(exc):
            raise exc

        def __raise_catch_cb(exc):
            try:
                raise exc
            except:
                pass
        event.timeout(0, __raise_cb, Exception())
        event.timeout(0, __raise_catch_cb, Exception())
        self.assertRaises(Exception, event.dispatch)

    def test_thread(self):
        self.call_back_ran_a = False
        self.call_back_ran_b = False

        def __time_cb(d):
            self.call_back_ran_a = True
            assert d['count'] == 3

        def __time_thread(count, d):
            self.call_back_ran_b = True
            for i in range(count):
                time.sleep(1)
                d['count'] += 1
        d = {'count': 0}
        _thread.start_new_thread(__time_thread, (3, d))
        event.timeout(4, __time_cb, d)
        event.dispatch()
        self.assertTrue(self.call_back_ran_a, 'call back a did not run')
        self.assertTrue(self.call_back_ran_b, 'call back b did not run')

if __name__ == '__main__':
    if input("run these tests with select registrar? (y/N)").lower().startswith("y"):
        rel.initialize(['select'], ['verbose', 'strict'])
    elif input('verbose mode? (y/N)').lower().startswith('y'):
        rel.initialize(options=['verbose'])
    unittest.main()



"""Defines the type of states that can be found in the  features provided by a Homely device."""
import datetime
from dataclasses import dataclass
from typing import Any, Optional

from dateutil.parser import parse


@dataclass
class State:
    feature_name: str

    @classmethod
    def create_from_rest_response(cls, data: dict) -> "State":
        raise NotImplementedError


def extract_value_and_last_updated(data: dict) -> tuple[Any, datetime.datetime]:
    timestamp = parse(data["lastUpdated"]) if data["lastUpdated"] is not None else None
    return data["value"], timestamp


@dataclass
class BasicAlarmState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "BasicAlarmState":
        my_data = data["features"]["alarm"]["states"]
        return BasicAlarmState(
            "alarm",
            *extract_value_and_last_updated(my_data["alarm"]),
            *extract_value_and_last_updated(my_data["tamper"]),
        )

    alarm: bool
    alarm_last_updated: datetime.datetime
    tamper: bool
    tamper_last_updated: datetime.datetime

    def __str__(self):
        return f"Alarm: {self.alarm} at {self.alarm_last_updated}, Tamper: {self.tamper} at {self.tamper_last_updated}"


@dataclass
class SmokeAlarmState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "SmokeAlarmState":
        my_data = data["features"]["alarm"]["states"]
        return SmokeAlarmState(
            "alarm",
            *extract_value_and_last_updated(my_data["fire"]),
        )

    fire: bool
    fire_last_updated: datetime.datetime

    def __str__(self):
        return f"Fire: {self.fire} at {self.fire_last_updated}"


@dataclass
class MotionSensorState(BasicAlarmState):
    sensitivity_level: Optional[float]
    sensitivity_level_last_updated: Optional[datetime.datetime]

    @classmethod
    def create_from_rest_response(cls, data: dict) -> "MotionSensorState":
        my_data = data["features"]["alarm"]["states"]
        return MotionSensorState(
            "alarm",
            *extract_value_and_last_updated(my_data["alarm"]),
            *extract_value_and_last_updated(my_data["tamper"]),
            *extract_value_and_last_updated(my_data["sensitivitylevel"]),
        )

    def __str__(self):
        return (
            super().__str__()
            + f", Sensitivity level: {self.sensitivity_level} at {self.sensitivity_level_last_updated}"
        )


@dataclass
class BatteryState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "BatteryState":
        my_data = data["features"]["battery"]["states"]
        return BatteryState(
            "battery",
            *extract_value_and_last_updated(my_data["low"]),
            *extract_value_and_last_updated(my_data["voltage"]),
            *extract_value_and_last_updated(my_data["defect"]) if "defect" in my_data else (None, None),
        )

    low: bool
    low_last_updated: datetime.datetime
    voltage: float
    voltage_last_updated: datetime.datetime
    defect: Optional[Any]
    defect_last_updated: datetime.datetime

    def __str__(self):
        return (
            f"Low: {self.low} at {self.low_last_updated}, Voltage: {self.voltage} at {self.voltage_last_updated}, "
            f"Defect: {self.defect} at {self.defect_last_updated}"
        )


@dataclass
class TemperatureState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "TemperatureState":
        my_data = data["features"]["temperature"]["states"]
        return TemperatureState(
            "temperature",
            *extract_value_and_last_updated(my_data["temperature"]),
        )

    temperature: float
    temperature_last_updated: datetime.datetime

    def __str__(self):
        return f"Temperature: {self.temperature} at {self.temperature_last_updated}"


@dataclass
class MeteringState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "MeteringState":
        my_data = data["features"]["metering"]["states"]
        return MeteringState(
            "metering",
            *extract_value_and_last_updated(my_data["summationdelivered"]),
            *extract_value_and_last_updated(my_data["summationreceived"]),
            *extract_value_and_last_updated(my_data["demand"]),
            *extract_value_and_last_updated(my_data["check"]),
        )

    summation_delivered: float
    summation_delivered_last_updated: datetime.datetime
    summation_received: float
    summation_received_last_update: datetime.datetime
    demand: float
    demand_last_updated: datetime.datetime
    check: bool
    check_last_updated: datetime.datetime

    def __str__(self):
        return (
            f"Summation delivered: {self.summation_delivered} at {self.summation_delivered_last_updated}, "
            f"Summation received: {self.summation_received} at {self.summation_received_last_update}, "
            f"Demand: {self.demand} at {self.demand_last_updated}, "
            f"Check: {self.check} at {self.check_last_updated}"
        )


@dataclass
class DiagnosticState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "DiagnosticState":
        my_data = data["features"]["diagnostic"]["states"]
        return DiagnosticState(
            "diagnostic",
            *extract_value_and_last_updated(my_data["networklinkaddress"]),
            *extract_value_and_last_updated(my_data["networklinkstrength"]),
        )

    network_link_address: str
    network_link_address_last_updated: datetime.datetime
    network_link_strength: float
    network_link_strength_last_updated: datetime.datetime

    def __str__(self):
        return (
            f"Network link address: {self.network_link_address} at {self.network_link_address_last_updated}, "
            f"Network link strength: {self.network_link_strength} at {self.network_link_strength_last_updated}"
        )



from collections import defaultdict

class Emitter(object):
    def __init__(self):
        self.handlers = defaultdict(list)

    def on(self, event, func=None):
        def set_event_handler(handler):
            self.handlers[event].append(handler)
            return handler

        if func:
            set_event_handler(func)
        else:
            return set_event_handler

    def off(self, event, func):
        self.handlers[event].remove(func)

    def once(self, event, func=None):
        def set_event_handler(handler):

            def wrapper(*args, **kwargs):
                self.off(event, wrapper)
                return handler(*args, **kwargs)

            self.on(event, wrapper)
            return handler

        if func:
            set_event_handler(func)
        else:
            return set_event_handler

    def emit(self, event, *args, **kwargs):
        for handler in self.handlers[event]:
            handler(*args, **kwargs)

    def removeAllListeners(self):
        self.handlers.clear()

